#!/usr/bin/env ruby
# encoding: utf-8
# test with ruby 1.8.7, 1.9.1, 1.9.3 on 2012-07-19
=begin
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2012 Kevin Redon <kevredon@mail.tsaitgaist.info>
=end
require 'socket'
require 'net/https'

# how to decode the terminal profile
# from ETSI TS 102 223 V11.0.0 §5.2
# each entry is for one byte
# each byte has a description
# each byte has multiple fields
# each field has a bit size
# each field has a description
TP = [
  {:desc => "Download", :fields => [
    {:bit_size => 1, :desc => "Profile download"},
    {:bit_size => 1, :desc => "3GPP: SMS-PP data download"},
    {:bit_size => 1, :desc => "3GPP: Cell Broadcast data download"},
    {:bit_size => 1, :desc => "Menu selection"},
    {:bit_size => 1, :desc => "3GPP: SMS-PP data download"},
    {:bit_size => 1, :desc => "Timer expiration"},
    {:bit_size => 1, :desc => "3GPP: USSD string data object support in Call Control by USIM"},
    {:bit_size => 1, :desc => "Call Control by NAA"}
  ]},
  {:desc => "Other", :fields => [
    {:bit_size => 1, :desc => "Command result"},
    {:bit_size => 1, :desc => "Call Control by NAA"},
    {:bit_size => 1, :desc => "Call Control by NAA"},
    {:bit_size => 1, :desc => "MO short message control support"},
    {:bit_size => 1, :desc => "Call Control by NAA"},
    {:bit_size => 1, :desc => "UCS2 Entry supported"},
    {:bit_size => 1, :desc => "UCS2 Display supported"},
    {:bit_size => 1, :desc => "Display Text"}
  ]},
  {:desc => "Proactive UICC", :fields => [
    {:bit_size => 1, :desc => "DISPLAY TEXT"},
    {:bit_size => 1, :desc => "GET INKEY"},
    {:bit_size => 1, :desc => "GET INPUT"},
    {:bit_size => 1, :desc => "MORE TIME"},
    {:bit_size => 1, :desc => "PLAY TONE"},
    {:bit_size => 1, :desc => "POLL INTERVAL"},
    {:bit_size => 1, :desc => "POLLING OFF"},
    {:bit_size => 1, :desc => "REFRESH"}
  ]},
  {:desc => "Proactive UICC", :fields => [
    {:bit_size => 1, :desc => "SELECT ITEM"},
    {:bit_size => 1, :desc => "3GPP: SEND SHORT MESSAGE with 3GPP-SMS-TPDU"},
    {:bit_size => 1, :desc => "3GPP: SEND SS"},
    {:bit_size => 1, :desc => "3GPP: SEND USSD"},
    {:bit_size => 1, :desc => "SET UP CALL"},
    {:bit_size => 1, :desc => "SET UP MENU"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (MCC, MNC, LAC, Cell ID & IMEI)"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (NMR)"}
  ]},
  {:desc => "Event driven information", :fields => [
    {:bit_size => 1, :desc => "Proactive UICC: SET UP EVENT LIST"},
    {:bit_size => 1, :desc => "MT call"},
    {:bit_size => 1, :desc => "Call connected"},
    {:bit_size => 1, :desc => "Call disconnected"},
    {:bit_size => 1, :desc => "Location status"},
    {:bit_size => 1, :desc => "User activity"},
    {:bit_size => 1, :desc => "Idle screen available"},
    {:bit_size => 1, :desc => "Card reader status"}
  ]},
  {:desc => "Event driven information extensions (for class a)", :fields => [
    {:bit_size => 1, :desc => "Language selection"},
    {:bit_size => 1, :desc => "Browser Termination"},
    {:bit_size => 1, :desc => "Data available"},
    {:bit_size => 1, :desc => "Channel status"},
    {:bit_size => 1, :desc => "Access Technology Change"},
    {:bit_size => 1, :desc => "Display parameters changed"},
    {:bit_size => 1, :desc => "Local Connection"},
    {:bit_size => 1, :desc => "Network Search Mode Change"}
  ]},
  {:desc => "Multiple card proactive commands (for class a) (Proactive UICC)", :fields => [
    {:bit_size => 1, :desc => "POWER ON CARD"},
    {:bit_size => 1, :desc => "POWER OFF CARD"},
    {:bit_size => 1, :desc => "PERFORM CARD APDU"},
    {:bit_size => 1, :desc => "GET READER STATUS (Card reader status)"},
    {:bit_size => 1, :desc => "GET READER STATUS (Card reader identifier)"}
  ]},
  {:desc => "Proactive UICC", :fields => [
    {:bit_size => 1, :desc => "TIMER MANAGEMENT (start, stop)"},
    {:bit_size => 1, :desc => "TIMER MANAGEMENT (get current value)"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (date, time and time zone)"},
    {:bit_size => 1, :desc => "GET INKEY"},
    {:bit_size => 1, :desc => "SET UP IDLE MODE TEXT"},
    {:bit_size => 1, :desc => "RUN AT COMMAND"},
    {:bit_size => 1, :desc => "SETUP CALL"},
    {:bit_size => 1, :desc => "Call Control by NAA"}
  ]},
  {:desc => "Proactive UICC", :fields => [
    {:bit_size => 1, :desc => "DISPLAY TEXT"},
    {:bit_size => 1, :desc => "SEND DTMF command"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (NMR)"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (language)"},
    {:bit_size => 1, :desc => "3GPP: PROVIDE LOCAL INFORMATION, Timing Advance"},
    {:bit_size => 1, :desc => "LANGUAGE NOTIFICATION"},
    {:bit_size => 1, :desc => "LAUNCH BROWSER"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (Access Technology)"}
  ]},
  {:desc => "Soft keys support (for class d)", :fields => [
    {:bit_size => 1, :desc => "Soft keys support for SELECT ITEM"},
    {:bit_size => 1, :desc => "Soft keys support for SET UP MENU"}
  ]},
  {:desc => "Soft keys information", :fields => [
    {:bit_size => 8, :desc => "Maximum number of soft keys available"}
  ]},
  {:desc => "Bearer Independent protocol proactive commands (for class e) (Proactive UICC)", :fields => [
    {:bit_size => 1, :desc => "OPEN CHANNEL"},
    {:bit_size => 1, :desc => "CLOSE CHANNEL"},
    {:bit_size => 1, :desc => "RECEIVE DATA"},
    {:bit_size => 1, :desc => "SEND DATA"},
    {:bit_size => 1, :desc => "GET CHANNEL STATUS"},
    {:bit_size => 1, :desc => "SERVICE SEARCH"},
    {:bit_size => 1, :desc => "GET SERVICE INFORMATION"},
    {:bit_size => 1, :desc => "DECLARE SERVICE"}
  ]},
  {:desc => "Bearer Independent protocol proactive commands (for class e)", :fields => [
    {:bit_size => 1, :desc => "CSD"},
    {:bit_size => 1, :desc => "GPRS"},
    {:bit_size => 1, :desc => "Bluetooth"},
    {:bit_size => 1, :desc => "IrDA"},
    {:bit_size => 1, :desc => "RS232"},
    {:bit_size => 3, :desc => "Number of channels supported by terminal"}
  ]},
  {:desc => "Screen height", :fields => [
    {:bit_size => 5, :desc => "Number of characters supported down the terminal display"},
    {:bit_size => 1, :desc => "No display capability"},
    {:bit_size => 1, :desc => "No keypad available"},
    {:bit_size => 1, :desc => "Screen Sizing Parameters supported"}
  ]},
  {:desc => "Screen width", :fields => [
    {:bit_size => 7, :desc => "Number of characters supported across the terminal display"},
    {:bit_size => 1, :desc => "Variable size fonts"}
  ]},
  {:desc => "Screen effects", :fields => [
    {:bit_size => 1, :desc => "Display can be resized"},
    {:bit_size => 1, :desc => "Text Wrapping supported"},
    {:bit_size => 1, :desc => "Text Scrolling supported"},
    {:bit_size => 1, :desc => "Text Attributes supported"},
    {:bit_size => 1, :desc => "RFU"},
    {:bit_size => 3, :desc => "Width reduction when in a menu"}
  ]},
  {:desc => "Bearer independent protocol supported transport interface/bearers (for class e)", :fields => [
    {:bit_size => 1, :desc => "TCP, UICC in client mode, remote connection"},
    {:bit_size => 1, :desc => "UDP, UICC in client mode, remote connection"},
    {:bit_size => 1, :desc => "TCP, UICC in server mode"},
    {:bit_size => 1, :desc => "TCP, UICC in client mode, local connection (i.e. class k is supported)"},
    {:bit_size => 1, :desc => "UDP, UICC in client mode, local connection (i.e. class k is supported)"},
    {:bit_size => 1, :desc => "Direct communication channel (i.e. class k is supported)"},
    {:bit_size => 1, :desc => "3GPP: E-UTRAN"},
    {:bit_size => 1, :desc => "3GPP: HSDPA"}
  ]},
  {:desc => "Proactive UICC", :fields => [
    {:bit_size => 1, :desc => "DISPLAY TEXT (Variable Time out)"},
    {:bit_size => 1, :desc => "GET INKEY (help is supported while waiting for immediate response or variable timeout)"},
    {:bit_size => 1, :desc => "USB (Bearer Independent protocol supported bearers, class e)"},
    {:bit_size => 1, :desc => "GET INKEY (Variable Timeout)"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (ESN)"},
    {:bit_size => 1, :desc => "Call control on GPRS"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (IMEISV)"},
    {:bit_size => 1, :desc => "PROVIDE LOCAL INFORMATION (Search Mode change)"}
  ]},
  {:desc => "reserved for TIA/EIA-136-C facilities", :fields => [
    {:bit_size => 4, :desc => "Protocol Version support"}
  ]},
  {:desc => "reserved for TIA/EIA/IS-820-A facilities", :fields => [] },
  {:desc => "Extended Launch Browser Capability (for class c)", :fields => [
    {:bit_size => 1, :desc => "WML"},
    {:bit_size => 1, :desc => "XHTML"},
    {:bit_size => 1, :desc => "HTML"},
    {:bit_size => 1, :desc => "CHTML"}
  ]},
  {:desc => "", :fields => [
    {:bit_size => 1, :desc => "3GPP: Support of UTRAN PS with extended parameters"},
    {:bit_size => 1, :desc => "Proactive UICC: PROVIDE LOCAL INFORMATION (battery state), (i.e. class g is supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: PLAY TONE (Melody tones and Themed tones supported)"},
    {:bit_size => 1, :desc => "Multi-media Calls in SET UP CALL (if class h supported)"},
    {:bit_size => 1, :desc => "3GPP: Toolkit-initiated GBA"},
    {:bit_size => 1, :desc => "Proactive UICC: RETRIEVE MULTIMEDIA MESSAGE (if class j is supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: SUBMIT MULTIMEDIA MESSAGE (if class j is supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: DISPLAY MULTIMEDIA MESSAGE (if class j is supported)"}
  ]},
  {:desc => "", :fields => [
    {:bit_size => 1, :desc => "Proactive UICC: SET FRAMES (i.e. class i is supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: GET FRAMES STATUS (i.e. class i is supported)"},
    {:bit_size => 1, :desc => "MMS notification download (if class j is supported)"},
    {:bit_size => 1, :desc => "Alpha Identifier in REFRESH command supported by terminal"},
    {:bit_size => 1, :desc => "3GPP: Geographical Location Reporting (if class n is supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: PROVIDE LOCAL INFORMATION (MEID)"},
    {:bit_size => 1, :desc => "Proactive UICC: PROVIDE LOCAL INFORMATION (NMR(UTRAN/E-UTRAN))"},
    {:bit_size => 1, :desc => "3GPP: USSD Data download and application mode"}
  ]},
  {:desc => "(for class i)", :fields => [
    {:bit_size => 4, :desc => "Maximum number of frames supported (including frames created in existing frames)"}
  ]},
  {:desc => "Event driven information extensions", :fields => [
    {:bit_size => 1, :desc => "Event: Browsing status"},
    {:bit_size => 1, :desc => "Event: MMS Transfer status (if class j is supported)"},
    {:bit_size => 1, :desc => "Event: Frame Information changed (i.e. class i is supported)"},
    {:bit_size => 1, :desc => "3GPP: Event: I-WLAN Access status (if class e is supported)"},
    {:bit_size => 1, :desc => "3GPP: Event Network Rejection"},
    {:bit_size => 1, :desc => "Event: HCI connectivity event (i.e. class m is supported)"},
    {:bit_size => 1, :desc => "3GPP: E-UTRAN support in Event Network Rejection)"},
    {:bit_size => 1, :desc => "Multiple access technologies supported in Event Access Technology Change and PROVIDE LOCAL INFORMATION"}
  ]},
  {:desc => "Event driven information extensions", :fields => [
    {:bit_size => 1, :desc => "Event : CSG Cell Selection (if class q is supported"},
    {:bit_size => 1, :desc => "Event: Contactless state request (if class r is supported)"},
    {:bit_size => 1, :desc => "Multiple access technologies supported in Event Access Technology Change and PROVIDE LOCAL INFORMATION"}
  ]},
  {:desc => "Event driven information extensions", :fields => [] },
  {:desc => "Text attributes", :fields => [
    {:bit_size => 1, :desc => "Alignment left supported by Terminal"},
    {:bit_size => 1, :desc => "Alignment centre supported by Terminal"},
    {:bit_size => 1, :desc => "Alignment right supported by Terminal"},
    {:bit_size => 1, :desc => "Font size normal supported by Terminal"},
    {:bit_size => 1, :desc => "Font size large supported by Terminal"},
    {:bit_size => 1, :desc => "Font size small supported by Terminal"}
  ]},
  {:desc => "Text attributes", :fields => [
    {:bit_size => 1, :desc => "Style normal supported by Terminal"},
    {:bit_size => 1, :desc => "Style bold supported by Terminal"},
    {:bit_size => 1, :desc => "Style italic supported by Terminal"},
    {:bit_size => 1, :desc => "Style underlined supported by Terminal"},
    {:bit_size => 1, :desc => "Style strikethrough supported by Terminal"},
    {:bit_size => 1, :desc => "Style text foreground colour supported by Terminal"},
    {:bit_size => 1, :desc => "Style text background colour supported by Terminal"}
  ]},
  {:desc => "", :fields => [
    {:bit_size => 1, :desc => "3GPP: I-WLAN bearer support (if class e is supported)"},
    {:bit_size => 1, :desc => "3GPP: Proactive UICC: PROVIDE LOCAL INFORMATION (WSID of the current I-WLAN connection)"},
    {:bit_size => 1, :desc => "TERMINAL APPLICATIONS (i.e. class k is supported)"},
    {:bit_size => 1, :desc => "3GPP: Steering of Roaming REFRESH support"},
    {:bit_size => 1, :desc => "Proactive UICC: ACTIVATE (i.e. class l is supported)"},
    {:bit_size => 1, :desc => "3GPP: Proactive UICC: GEOGRAPHICAL LOCATION REQUEST (if class n is supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: PROVIDE LOCAL INFORMATION (Broadcast Network Information) (i.e. class o is supported)"},
    {:bit_size => 1, :desc => "3GPP: Steering of Roaming for I-WLAN REFRESH support"}
  ]},
  {:desc => "", :fields => [
    {:bit_size => 1, :desc => "Proactive UICC: Contactless State Changed (if class r is supported)"},
    {:bit_size => 1, :desc => "3GPP: Support of CSG cell discovery (if class q is supported)"},
    {:bit_size => 1, :desc => "Confirmation parameters supported for OPEN CHANNEL in Terminal Server Mode"},
    {:bit_size => 1, :desc => "3GPP: Communication Control for IMS"},
    {:bit_size => 1, :desc => "Support of CAT over the modem interface (if class s is supported)"},
    {:bit_size => 1, :desc => "3GPP Support for Incoming IMS Data event (if classes e and t are supported)"},
    {:bit_size => 1, :desc => "3GPP Support for IMS Registration event (if classes e and t are supported)"},
    {:bit_size => 1, :desc => "Proactive UICC: Profile Container, Envelope Container, COMMAND CONTAINER and ENCAPSULATED SESSION CONTROL (if class u is supported)"}
  ]},
  {:desc => "", :fields => [
    {:bit_size => 1, :desc => "3GPP: Support of IMS as a bearer for BIP (if classes e and t are supported)"}
  ]}
]

class String
  def to_hex_disp
    to_return = ""
    each_byte do |b|
      to_return += b.to_s(16).rjust(2,"0")
      to_return += " "
    end
    return to_return[0..-2].upcase
  end

  def to_hex
    to_return = ""
    each_byte do |b|
      to_return += b.to_s(16).rjust(2,"0")
    end
    #to_return = "0x"+to_return
    return to_return.downcase
  end
  
  # convert a hexadecimal string into binary array
  def hex2arr
    arr = []
    (self.length/2).times do |i|
      arr << self[i*2,2].to_i(16)
    end
    return arr
  end
end

class Array
  def to_hex_disp
    to_return = ""
    each do |b|
      to_return += b.to_s(16).rjust(2,"0")
      to_return += " "
    end
    return to_return[0..-2].upcase
  end

  def to_hex
    to_return = ""
    each do |b|
      to_return += b.to_s(16).rjust(2,"0")
    end
    #to_return = "0x"+to_return
    return to_return.downcase
  end
end

# decode terminal profile (from ETSI TS 102 223 V11.0.0 §5.2)
# tp is treated as byte array
def decode_terminal_profile(tp)
  tp.each_index do |i|
    puts TP[i][:desc]+ " (0x#{tp[i].to_s(16).rjust(2,'0')})"
    bit = 0
    TP[i][:fields].each do |field|
      print "\t- "
      value = (tp[i]>>bit)&((1<<field[:bit_size])-1)
      if field[:bit_size] == 1 then
        puts field[:desc] + ": " + ( value==1 ? "" : "not " ) + "supported"
      else
        puts field[:desc] + ": #{value}"
      end
      bit += field[:bit_size]
    end
  end
end

# decode terminal profile if given as parameter
if ARGV.size>0 then
  raise "bad terminal profile" if ARGV[0].length%2!=0 or ARGV[0]=~/[^a-fA-F0-9]/
  tp = ARGV[0].hex2arr
  decode_terminal_profile(tp)
  exit 0
end

# create UDP socket to receive GSMTAP messages
socket = UDPSocket.new(Socket::AF_INET6)
socket.bind("::", 4729)

@end = false
until @end do
  activity = IO.select([socket])
  begin
    # read a complete udp packet (as byte array)
    input = activity[0][0].readpartial(65507).unpack("C*")
    # verify it's a SIM GSMTAP packet
    next unless input.length>16 # GSMTAP header
    next unless input[1] == 0x4 # header length
    next unless input[2] == 0x4 # SIM type
    # verify apdu
    apdu = input[16..-1]
    # discard ATR
    next if apdu[0] == 0x3b or apdu[0] == 0x3f
    # verify it's a terminal profile APDU (from ETSI TS 102 221 §10.1.2)
    next unless apdu[0,4]==[0xA0,0x10,0x00,0x00] or apdu[0,4]==[0x80,0x10,0x00,0x00]
    puts "found terminal profile command: " + apdu.to_hex_disp
    # decode it
    puts "decoding terminal profile:"
    tp = apdu[5,apdu[4]]
    decode_terminal_profile(tp)
    puts
    # get phone information
    brand = nil
    until brand do
      puts "enter phone brand:"
      brand = gets.chomp
      brand = nil if brand.length==0
    end
    model = nil
    until model do
      puts "enter phone model:"
      model = gets.chomp
      model = nil if model.length==0
    end
    tac = nil
    until tac do
      puts "enter phone TAC (first 8 digits of IMEI):"
      tac = gets.chomp
      tac = nil unless tac=~/^[0-9]{8}$/
    end
    info = nil
    until info do
      puts "enter additional information (baseband version, …):"
      info = gets.chomp
    end
    # send data
    url = "terminal-profile.tsaitgaist.info"
    puts "following information will be send to #{url}:"
    puts "- phone brand: #{brand}"
    puts "- phone model: #{model}"
    puts "- phone TAC: #{tac}"
    puts "- additional information: #{info}"
    puts "- terminal profile: #{tp.to_hex_disp}"
    puts "These information will available under CC-BY-SA. Do you agree [Y/n]?"
    agree = gets.chomp.downcase
    if agree.length==0 or agree=="y" then
      https = Net::HTTP.new(url, Net::HTTP.https_default_port)
      https.use_ssl = true
      https.verify_mode = OpenSSL::SSL::VERIFY_NONE
      res = https.post('/post.php', "brand=#{brand}&model=#{model}&tac=#{tac}&info=#{info}&tp=#{tp.to_hex}")
      if res.code=="200" then
        puts "entry sent successfully"
        puts "thank you for your contribution"
      else
        puts "sending entry failed"
      end
    else
      puts "aborting. the information will not be uploaded"
    end
    @end = true
  rescue EOFError
    puts "connection closed"
  end
end

socket.close
